import json
import os
import shutil

import msgpack

from flask import request, jsonify, Response, Blueprint

from ....configs.config import config
from ....core_components.safe_dict import SafeDict


database_ops = Blueprint('database_ops', __name__)


tasks = SafeDict()

root_path = config.LYNSE_DEFAULT_ROOT_PATH


@database_ops.before_request
def handle_msgpack_input():
    if request.headers.get('Content-Type') == 'application/msgpack':
        data = msgpack.unpackb(request.get_data(), raw=False)
        request.decoded_data = data


@database_ops.route('/', methods=['GET'])
def index():
    return Response(json.dumps({'status': 'success', 'message': 'LynseDB HTTP API'}), mimetype='application/json')


@database_ops.route('/drop_database', methods=['POST'])
def drop_database():
    data = request.json
    if not data or 'database_name' not in data:
        return jsonify({'error': 'No data provided or missing database_name'}), 400

    try:
        shutil.rmtree(root_path / data['database_name'])
        return jsonify({'status': 'success'})
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@database_ops.route('/database_exists', methods=['POST'])
def database_exists():
    """Check if the database exists.

    Returns:
        dict: The status of the operation.
    """
    data = request.json
    try:
        exists = (root_path / data['database_name']).exists()
        return Response(json.dumps({
            'status': 'success', 'params': {
                'exists': True if exists else False
            }
        },
            sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@database_ops.route('/set_environment', methods=['POST'])
def set_environment():
    """Set the environment variables.

    Returns:
        dict: The status of the operation.
    """
    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        for key, value in data.items():
            os.environ[key] = value
        return Response(json.dumps({'status': 'success', 'params': data}, sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@database_ops.route('/get_environment', methods=['GET'])
def get_environment():
    """Get the environment variables.

    Returns:
        dict: The status of the operation.
    """
    env_list = ['LYNSE_LOG_LEVEL', 'LYNSE_LOG_PATH', 'LYNSE_TRUNCATE_LOG', 'LYNSE_LOG_WITH_TIME',
                'LYNSE_KMEANS_EPOCHS', 'LYNSE_SEARCH_CACHE_SIZE']

    params = {key: eval("global_config.key") for key in env_list}
    try:
        return Response(json.dumps({'status': 'success', 'params': params}, sort_keys=False),
                        mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@database_ops.route('/list_databases', methods=['GET'])
def list_databases():
    """List all databases.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.database_manager import DatabaseManager

    try:
        data_manager = DatabaseManager(config.LYNSE_DEFAULT_ROOT_PATH)
        databases = data_manager.list_database()
        return Response(json.dumps({'status': 'success', 'params': {'databases': databases}}, sort_keys=False),
                        mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@database_ops.route('/delete_database', methods=['POST'])
def delete_database():
    """Delete a database.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.database_manager import DatabaseManager

    data = request.json
    try:
        data_manager = DatabaseManager(config.LYNSE_DEFAULT_ROOT_PATH)
        data_manager.delete(data['database_name'])
        return Response(json.dumps({'status': 'success', 'params': {'database_name': data['database_name']}},
                                   sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@database_ops.route('/create_database', methods=['POST'])
def create_database():
    """Create a database.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.database_manager import DatabaseManager
    from ....api.native_api.high_level import LocalClient

    data = request.json
    try:
        data_manager = DatabaseManager(config.LYNSE_DEFAULT_ROOT_PATH)
        data_manager.register(data['database_name'])

        if data['drop_if_exists']:
            LocalClient(root_path=root_path / data['database_name']).drop_database()

        LocalClient(root_path=root_path / data['database_name'])
        return Response(json.dumps({'status': 'success', 'params': {'database_name': data['database_name']}},
                                   sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500



# The address of the docker container running the service is http://localhost:5403
import time
import numpy as np

from test import VectorDBClient
from test import HTTPClient, Collection


def test_initialization():
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=True)

    assert isinstance(db, HTTPClient)


def test_create_collection():
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=True)
    collection = db.require_collection('test_collection', dim=4, drop_if_exists=True)
    assert isinstance(collection, Collection)
    assert collection._collection_name == 'test_collection'


def test_add_item():
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=False)
    collection = db.require_collection('test_collection', dim=4, drop_if_exists=True)
    item = {
        "vector": [0.01, 0.34, 0.74, 0.31],
        "id": 1,
        "field": {
            'field': 'test_1',
            'order': 0
        }
    }

    with collection.insert_session():
        id = collection.add_item(**item)
    assert id == 1


def test_bulk_add_items():
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=False)
    collection = db.require_collection('test_collection', dim=4, drop_if_exists=True)
    time.sleep(3)
    with collection.insert_session():
        ids = collection.bulk_add_items([([0.01, 0.34, 0.74, 0.31], 1, {'field': 'test_1', 'order': 0}),
                                            ([0.36, 0.43, 0.56, 0.12], 2, {'field': 'test_1', 'order': 1}),
                                            ([0.03, 0.04, 0.10, 0.51], 3, {'field': 'test_2', 'order': 2}),
                                            ([0.11, 0.44, 0.23, 0.24], 4, {'field': 'test_2', 'order': 3}),
                                            ([0.91, 0.43, 0.44, 0.67], 5, {'field': 'test_2', 'order': 4}),
                                            ([0.92, 0.12, 0.56, 0.19], 6, {'field': 'test_3', 'order': 5}),
                                            ([0.18, 0.34, 0.56, 0.71], 7, {'field': 'test_1', 'order': 6}),
                                            ([0.01, 0.33, 0.14, 0.31], 8, {'field': 'test_2', 'order': 7}),
                                            ([0.71, 0.75, 0.91, 0.82], 9, {'field': 'test_3', 'order': 8}),
                                            ([0.75, 0.44, 0.38, 0.75], 10, {'field': 'test_1', 'order': 9})])
    assert ids == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]


def test_query():
    import operator

    from lynse.core_components.fields_cache.filter import Filter, FieldCondition, MatchField, MatchID

    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=False)
    collection = db.get_collection('test_collection')
    collection.build_index("Flat-Cos")
    ids, scores, fields = collection.search(
        vector=[0.36, 0.43, 0.56, 0.12],
        k=10,
        search_filter=Filter(
            must=[
                FieldCondition(key='field', matcher=MatchField('test_1')),  # Support for filtering fields
            ],
            any=[
                FieldCondition(key='order', matcher=MatchField(8, comparator=operator.ge)),
                FieldCondition(key=":match_id:", matcher=MatchID([1, 2, 3, 4, 5])),  # Support for filtering IDs
            ]
        )
    )
    print(ids)
    assert np.array_equal(ids, np.array([2, 1, 10]))


def test_drop_collection():
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=False)
    db.drop_collection('test_collection')
    print(db.show_collections())
    assert 'test_collection' not in db.show_collections()



import concurrent.futures
import time

from test import VectorDBClient, HTTPClient


def test_initialization():
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=True)

    assert isinstance(db, HTTPClient)


def test_create_collection():
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=True)
    collection = db.require_collection('test_collection', dim=4, drop_if_exists=True)
    assert collection._collection_name == 'test_collection'


def test_add_item():
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=True)
    collection = db.require_collection('test_collection', dim=4, drop_if_exists=True)
    time.sleep(3)
    with collection.insert_session():
        id = collection.add_item([0.01, 0.34, 0.74, 0.31], id=1, field={'name': 'John Doe'})
        assert id == 1
        id = collection.add_item([0.36, 0.43, 0.56, 0.12], id=2, field={'name': 'Jane Doe'})
        assert id == 2
        id = collection.add_item([0.03, 0.04, 0.10, 0.51], id=3, field={'name': 'John Smith'})
        assert id == 3
        id = collection.add_item([0.11, 0.44, 0.23, 0.24], id=4, field={'name': 'Jane Smith'})
        assert id == 4

    assert collection.shape == (4, 4)


def add_item(id, vector, field):
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=False)
    collection = db.get_collection('test_collection')
    with collection.insert_session() as session:
        session.add_item(vector, id=id, field=field, buffer_size=True)


def bulk_add_items(items):
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=False)
    collection = db.get_collection('test_collection')
    collection.bulk_add_items(items)
    collection.commit()


def test_multi_users_bulk_add_items():
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=True)
    collection = db.require_collection('test_collection', dim=4, drop_if_exists=True)

    items = [
        ([1, 2, 3, 4], 1, {'name': 'John Doe'}),
        ([5, 6, 7, 8], 2, {'name': 'Jane Doe'}),
        ([9, 10, 11, 12], 3, {'name': 'John Smith'}),
        ([13, 14, 15, 16], 4, {'name': 'Jane Smith'})
    ]

    with concurrent.futures.ProcessPoolExecutor() as executor:
        futures = [executor.submit(bulk_add_items, items),
                   executor.submit(bulk_add_items, items),
                   executor.submit(bulk_add_items, items),
                   executor.submit(bulk_add_items, items)]

        concurrent.futures.wait(futures)

    assert collection.shape == (4, 4)


def test_multi_thread_bulk_add_items():
    client = VectorDBClient('http://localhost:7637')
    db = client.create_database('test_db', drop_if_exists=True)
    collection = db.require_collection('test_collection', dim=4, drop_if_exists=True)

    items = [
        ([1, 2, 3, 4], 1, {'name': 'John Doe'}),
        ([5, 6, 7, 8], 2, {'name': 'Jane Doe'}),
        ([9, 10, 11, 12], 3, {'name': 'John Smith'}),
        ([13, 14, 15, 16], 4, {'name': 'Jane Smith'})
    ]

    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = [executor.submit(bulk_add_items, items),
                   executor.submit(bulk_add_items, items),
                   executor.submit(bulk_add_items, items),
                   executor.submit(bulk_add_items, items)]

        concurrent.futures.wait(futures)

    assert collection.shape == (4, 4)



from itertools import chain

import numpy as np
import msgpack
import zlib


class IVFIndex:
    """
    Inverted File Index for storing the mapping between external IDs and file names.
    """

    def __init__(self):
        """
        Initialize the IVFIndex.
        """
        self.file_names = {}
        self.file_name_to_id = {}
        self.index = {}
        self.file_counter = 0
        self.last_external_id = 0

    def _get_file_id(self, file_name):
        """
        Get the file ID by file name.

        Parameters:
            file_name (str): The file name.

        Returns:
            int: The file ID.
        """
        if file_name in self.file_name_to_id:
            return self.file_name_to_id[file_name]

        self.file_counter += 1
        self.file_names[self.file_counter] = file_name
        self.file_name_to_id[file_name] = self.file_counter
        return self.file_counter

    def add_entry(self, cluster_id, file_name, external_id):
        """
        Add an entry to the specified cluster.

        Parameters:
            cluster_id (int): The cluster ID.
            file_name (str): The file name.
            external_id (int): The external ID.

        Returns:
            None
        """
        cluster_id = int(cluster_id)
        external_id = int(external_id)

        file_id = self._get_file_id(file_name)

        if cluster_id not in self.index:
            self.index[cluster_id] = {}

        if file_id not in self.index[cluster_id]:
            self.index[cluster_id][file_id] = set()

        # Check if the entry already exists to avoid duplicates
        if external_id in self.index[cluster_id][file_id]:
            return  # Entry already exists, do nothing

        self.index[cluster_id][file_id].add(external_id)
        self.last_external_id = external_id

    def get_entries(self, cluster_id):
        """
        Get all entries in the specified cluster.

        Parameters:
            cluster_id (int): The cluster ID.

        Returns:
            Dict: The dictionary of entries.
        """
        if cluster_id not in self.index:
            return {}

        file_entries = {}
        for file_id, external_ids in self.index[cluster_id].items():
            file_name = self.file_names[file_id]
            file_entries[file_name] = np.array(list(external_ids))

        return file_entries

    def remove_entry(self, cluster_id, external_id):
        """
        Remove the specified entry from the cluster.

        Parameters:
            cluster_id (int): The cluster ID.
            external_id (int): The external ID.

        Returns:
            bool: Whether the entry is successfully removed.
        """
        if cluster_id in self.index:
            for file_id, external_ids in self.index[cluster_id].items():
                if external_id in external_ids:
                    external_ids.remove(external_id)
                    if not external_ids:
                        del self.index[cluster_id][file_id]
                    return True
        return False

    def update_entry(self, cluster_id, external_id, new_file_name=None):
        """
        Update the specified entry in the cluster.

        Parameters:
            cluster_id (int): The cluster ID.
            external_id (int): The external ID.
            new_file_name (str): The new file name.

        Returns:
            bool: Whether the entry is successfully updated.
        """
        cluster_id = int(cluster_id)
        external_id = int(external_id)

        if cluster_id in self.index:
            for file_id, external_ids in self.index[cluster_id].items():
                if external_id in external_ids:
                    if new_file_name is not None:
                        new_file_id = self._get_file_id(new_file_name)
                        external_ids.remove(external_id)
                        if not external_ids:
                            del self.index[cluster_id][file_id]
                        if new_file_id not in self.index[cluster_id]:
                            self.index[cluster_id][new_file_id] = set()
                        self.index[cluster_id][new_file_id].add(external_id)
                    return True
        return False

    def clear_cluster(self, cluster_id):
        """
        Clear all entries in the specified cluster.

        Parameters:
            cluster_id (int): The cluster ID.

        Returns:
            None
        """
        if cluster_id in self.index:
            del self.index[cluster_id]

    def clear_all(self):
        """
        Clear all entries in the IVFIndex.
        """
        self.index.clear()
        self.file_names.clear()
        self.file_name_to_id.clear()
        self.file_counter = 0

    def save(self, file_path):
        """
        Save the IVFIndex to a file.

        Parameters:
            file_path (str or Pathlike): The file path.

        Returns:
            None
        """
        data = {
            'file_names': self.file_names,
            'index': {cluster_id: {file_id: list(external_ids) for file_id, external_ids in file_dict.items()}
                      for cluster_id, file_dict in self.index.items()},
            'file_counter': self.file_counter
        }
        packed_data = msgpack.packb(data)
        compressed_data = zlib.compress(packed_data)
        with open(file_path, 'wb') as f:
            f.write(compressed_data)

    def load(self, file_path):
        """
        Load the IVFIndex from a file.

        Parameters:
            file_path (str or Pathlike): The file path.

        Returns:
            IVFIndex: The IVFIndex object.
        """
        with open(file_path, 'rb') as f:
            compressed_data = f.read()
            packed_data = zlib.decompress(compressed_data)
            data = msgpack.unpackb(packed_data, strict_map_key=False)
            self.file_names = data['file_names']
            self.index = {cluster_id: {file_id: set(external_ids) for file_id, external_ids in file_dict.items()}
                          for cluster_id, file_dict in data['index'].items()}
            self.file_counter = data['file_counter']

        # Rebuild the reverse mapping
        self.file_name_to_id = {name: id for id, name in self.file_names.items()}

        return self

    def all_external_ids(self):
        """
        Get all external IDs in the IVFIndex.

        Returns:
            List: The list of external IDs.
        """
        return tuple(
            chain.from_iterable(
                external_ids for file_dict in self.index.values() for external_ids in file_dict.values()
            )
        )

    def __len__(self):
        return len(self.all_external_ids())



from sklearn.cluster import MiniBatchKMeans
import numpy as np


class BatchKMeans:
    def __init__(self, n_clusters, random_state=42, epochs=100, batch_size=1024):
        self.model = MiniBatchKMeans(n_clusters=n_clusters, random_state=random_state, max_iter=epochs,
                                     batch_size=batch_size, n_init='auto', reassignment_ratio=0)

        self.n_clusters = n_clusters
        self.epochs = epochs
        self.batch_size = batch_size

        self.fitted = False

    def fit(self, X: np.ndarray):
        if X.dtype != np.float32:
            X = X.astype(np.float32)

        self.model.fit(X)
        self.fitted = True

        return self

    def partial_fit(self, X: np.ndarray):
        if X.dtype != np.float32:
            X = X.astype(np.float32)

        self.model.partial_fit(X)
        self.fitted = True

        return self

    def predict(self, X: np.ndarray):
        if X.dtype != np.float32:
            X = X.astype(np.float32)

        if not self.fitted:
            raise ValueError('The model is not fitted yet. Please call the fit method first.')

        return self.model.predict(X)

    def save(self, filename):
        import cloudpickle

        with open(filename, 'wb') as f:
            cloudpickle.dump([self.model, self.n_clusters, self.epochs, self.batch_size], f)

    def load(self, filename):
        import cloudpickle

        self.model, self.n_clusters, self.epochs, self.batch_size = cloudpickle.load(open(filename, 'rb'))
        self.fitted = True

        return self

    @property
    def cluster_centers_(self):
        if not self.fitted:
            raise ValueError('The model is not fitted yet. Please call the fit method first.')
        return self.model.cluster_centers_

    @property
    def labels_(self):
        if not self.fitted:
            raise ValueError('The model is not fitted yet. Please call the fit method first.')
        return self.model.labels_



import os
import msgpack
import tempfile
import mmap
from typing import Dict
from pathlib import Path

from ...core_components.fields_cache.index import Index
from ...core_components.id_checker import IDChecker


class FieldsStorage:
    RECORD_HEADER_SIZE = 4

    def __init__(self, filepath=None, as_temp_file=False):
        if filepath is None and not as_temp_file:
            raise ValueError("You must provide a filepath or set as_temp_file to True.")
        if as_temp_file:
            self.filepath = Path(tempfile.mkstemp()[1])
        else:
            self.filepath = Path(filepath)

        self.memory_store = {}
        self.id_mapping = {}
        self.external_to_internal = {}
        self.current_internal_id = 0

        self.filter_path = self.filepath.with_suffix('.filter')
        self.index_path = self.filepath.with_suffix('.index')
        self._initialize_id_checker()

        self._load_mappings()

        self.index = Index()
        self._load_index()

        self.max_entries = 100000
        self.use_index = False

        self.all_fields = {}
        self._load_all_fields()

    def _initialize_id_checker(self):
        self.id_filter = IDChecker()

        if self.filter_path.exists():
            self.id_filter.from_file(self.filter_path)
        else:
            if self.filepath.exists() and self.filepath.stat().st_size > 0:
                for external_id, _ in self.retrieve_all():
                    self.id_filter.add(external_id)
                self.id_filter.to_file(self.filter_path)

    def _load_mappings(self):
        id_mapping_path = self.filepath.with_suffix('.idmap')
        external_to_internal_path = self.filepath.with_suffix('.ext2int')
        if id_mapping_path.exists():
            with open(id_mapping_path, 'rb') as f:
                unpacked = msgpack.unpackb(f.read(), raw=False, strict_map_key=False)
                self.id_mapping.update(unpacked)
            if self.id_mapping:
                self.current_internal_id = max(self.id_mapping.keys()) + 1
        if external_to_internal_path.exists():
            with open(external_to_internal_path, 'rb') as f:
                unpacked = msgpack.unpackb(f.read(), raw=False, strict_map_key=False)
                self.external_to_internal.update(unpacked)

    def _load_index(self):
        if self.index_path.exists():
            self.index.load(self.index_path)
            self.use_index = True

    def _load_all_fields(self):
        if self.filepath.exists() and self.filepath.stat().st_size > 0:
            for _, data in self.retrieve_all():
                for field, value in data.items():
                    if field not in self.all_fields:
                        self.all_fields[field] = type(value).__name__

    def auto_commit(self):
        if self.memory_store:
            self.commit()

    def store(self, data: dict, external_id: int):
        """
        Store a record in the cache.

        Parameters:
            data: dict
                The record to store.
            external_id: int
                The external ID of the record.

        Returns:
            int: The internal ID of the record.
        """
        if not isinstance(data, dict):
            raise ValueError("Only dictionaries are allowed as data.")

        if external_id in self.id_filter:
            raise ValueError(f"external_id {external_id} already exists.")
        self.id_filter.add(external_id)

        data = {":id:": external_id, **data}
        internal_id = self.current_internal_id
        self.memory_store[internal_id] = data
        self.external_to_internal[external_id] = internal_id
        self.current_internal_id += 1

        for field, value in data.items():
            self.all_fields[field] = type(value).__name__

        if self.use_index:
            self.index.insert(data, internal_id)

        if len(self.memory_store) >= self.max_entries:
            self._flush_to_disk()

        return internal_id

    def _flush_to_disk(self):
        if not self.memory_store:
            return

        with open(self.filepath, 'ab') as f:
            for internal_id, data in self.memory_store.items():
                packed_data = msgpack.packb({internal_id: data})
                record_size = len(packed_data)
                offset = f.tell()
                f.write(record_size.to_bytes(self.RECORD_HEADER_SIZE, 'big'))
                f.write(packed_data)
                self.id_mapping[internal_id] = offset
        self.memory_store = {}

        id_mapping_path = self.filepath.with_suffix('.idmap')
        with open(id_mapping_path, 'wb') as f:
            packed_mapping = msgpack.packb(self.id_mapping)
            f.write(packed_mapping)

        external_to_internal_path = self.filepath.with_suffix('.ext2int')
        with open(external_to_internal_path, 'wb') as f:
            packed_mapping = msgpack.packb(self.external_to_internal)
            f.write(packed_mapping)

        self.id_filter.to_file(self.filter_path)

        # save field list
        fields_path = self.filepath.with_suffix('.fields')
        with open(fields_path, 'wb') as f:
            packed_fields = msgpack.packb(list(self.all_fields))
            f.write(packed_fields)

        if self.use_index:
            self.index.save(self.index_path)

    def retrieve_all(self, include_external_id=True):
        self.auto_commit()

        if not os.path.exists(self.filepath):
            return

        with open(self.filepath, 'rb') as f:
            while True:
                record_size_bytes = f.read(self.RECORD_HEADER_SIZE)
                if not record_size_bytes:
                    break
                record_size = int.from_bytes(record_size_bytes, 'big')
                chunk = f.read(record_size)
                unpacked = msgpack.unpackb(chunk, raw=False, strict_map_key=False)
                for internal_id, data in unpacked.items():
                    external_id = data.get(':id:')
                    if not include_external_id:
                        del data[':id:']
                    yield external_id, data

    def retrieve_by_external_id(self, external_ids, include_external_id=True):
        self.auto_commit()

        if isinstance(external_ids, int):
            external_ids = [external_ids]

        results = []
        with open(self.filepath, 'rb') as f:
            with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mm:
                for external_id in external_ids:
                    if external_id in self.external_to_internal:
                        internal_id = self.external_to_internal[external_id]
                        offset = self.id_mapping[internal_id]
                        mm.seek(offset)
                        record_size = int.from_bytes(mm.read(self.RECORD_HEADER_SIZE), 'big')
                        chunk = mm.read(record_size)
                        unpacked = msgpack.unpackb(chunk, raw=False, strict_map_key=False)
                        if internal_id in unpacked:
                            if not include_external_id:
                                del unpacked[internal_id][':id:']
                            results.append(unpacked[internal_id])
        return results

    def delete(self):
        if self.filepath and self.filepath.exists():
            self.filepath.unlink()
        id_mapping_path = self.filepath.with_suffix('.idmap')
        if id_mapping_path.exists():
            id_mapping_path.unlink()
        external_to_internal_path = self.filepath.with_suffix('.ext2int')
        if external_to_internal_path.exists():
            external_to_internal_path.unlink()
        if self.filter_path.exists():
            self.filter_path.unlink()
        if self.index_path.exists():
            self.index_path.unlink()

        fields_path = self.filepath.with_suffix('.fields')
        if fields_path.exists():
            fields_path.unlink()

        self.memory_store.clear()
        self.id_mapping.clear()
        self.external_to_internal.clear()
        self.filepath = None
        self.memory_store = None
        self.id_mapping = None
        self.current_internal_id = None
        self.max_entries = None
        self.use_index = False
        self.index = None
        self.all_fields = set()

    def build_index(self, schema: Dict[str, type], rebuild_if_exists=False):
        self.auto_commit()

        # pre-select data to check the schema
        _, pre_select_data = next(self.retrieve_all())

        for index_name, index_type in schema.items():
            # dtype check
            if not isinstance(pre_select_data.get(index_name), index_type):
                raise TypeError(f"Field {index_name} is not of type {index_type}")

            self.index.add_index(index_name, index_type, rebuild_if_exists)
        for external_id, data in self.retrieve_all():
            self.index.insert(data, external_id)
        self.use_index = True
        self.index.save(self.index_path)

    def remove_index(self, field_name: str):
        self.index.remove_index(field_name)
        if not self.index.indices:
            if os.path.exists(self.index_path):
                os.remove(self.index_path)

            self.index = Index()
            self.use_index = False
        else:
            self.index.save(self.index_path)

    def commit(self):
        self._flush_to_disk()

    def list_fields(self):
        """
        Return all field names and their types.

        Returns:
            dict: A dictionary with field names as keys and field types as values.
        """
        def _warp_field_name(field):
            return f":{field}:" if field not in [':id:', ''] else field

        return {_warp_field_name(field): dtype for field, dtype in self.all_fields.copy().items()}



import httpx


class Poster:
    def __init__(self, retries=3, timeout=None, http2=False):
        """
        A class for making HTTP requests.

        Parameters:
            retries (int): The number of times to retry the request.
            timeout (int or None): The request timeout in seconds.
            http2 (bool): Whether to use HTTP/2.
        """
        self.retries = retries
        self.timeout = timeout
        self.http2 = http2
        transport = httpx.HTTPTransport(retries=retries, http2=http2)
        self.session = httpx.Client(transport=transport, timeout=timeout)

    def post(self, url, data=None, headers=None, content=None, json=None):
        """
        Make a POST request.

        Parameters:
            url (str): The URL to request.
            data (dict or None): The data to send.
            headers (dict or None): The headers to send.
            content (str or None or btypes): The content to send.
            json (dict or None): The JSON data to send.

        Returns:
            httpx.Response: The response object.
        """
        return self.session.post(url, data=data, headers=headers, content=content, json=json)

    def get(self, url, headers=None):
        """
        Make a GET request.

        Parameters:
            url (str): The URL to request.
            headers (dict or None): The headers to send.

        Returns:
            httpx.Response: The response object.
        """
        return self.session.get(url, headers=headers)




__version__ = '0.1.5'

from pathlib import Path
from typing import Union

from .api.http_api.http_api.app import launch_in_jupyter
from .core_components import fields_cache as field_models
from .configs.config import generate_config_file, load_config_file

generate_config_file(), load_config_file()


class _InstanceDistributor:
    from pathlib import Path
    from typing import Union

    def __new__(cls, root_path: Union[str, Path, None], database_name: str):
        from pathlib import Path
        from .api.native_api.high_level import LocalClient
        from .api.http_api.client_api import HTTPClient
        from .configs.config import config

        if isinstance(root_path, Path):
            root_path = root_path.as_posix()

        if root_path is not None and (root_path.startswith('http://') or root_path.startswith('https://')):
            instance = HTTPClient(uri=root_path, database_name=database_name)
        else:
            if root_path is None:
                native_api_root_path = config.LYNSE_DEFAULT_ROOT_PATH
            else:
                # new features, root_path can be specified directly
                native_api_root_path = Path(root_path)

            instance = LocalClient(native_api_root_path / database_name)

        return instance


class VectorDBClient:
    """
    This class determines whether it is local or remote based on the URI, thereby implementing the distribution of local and remote client communication.
    The data storage location is also determined accordingly.

    - **When `uri` is None**: The local client is used by default, and the data is stored under the default root path.
    - **When `uri` is a local path**: The local client is used, and the data is stored under the specified path.
    - **When `uri` is a remote URL**: The remote client is used, and the data is stored on the remote server.
    """
    def __init__(self, uri: Union[str, None, Path] = None):
        """
        Initialize the LynseDB client.

        Parameters:
            uri (Pathlike or str or None): The URI of the LynseDB server. It can be either a local path or a remote URL.

               - If it is a remote URL, the client will use the HTTP API.
               - If it is a local path, the client will use the native API.
                    The path refers to the root path of the LynseDB storage.
               - If set to None, the client will use the native API,
                    and the database will be stored in the default root path,
                    when you need to change the default root path,
                    you can set the environment variable LYNSE_DEFAULT_ROOT_PATH or change the config file.

        """
        self._is_remote = uri is not None and (uri.startswith('http://') or uri.startswith('https://'))

        if self._is_remote:
            from spinesUtils.asserts import raise_if
            import httpx

            raise_if(ValueError, not isinstance(uri, (str, Path)),
                     'uri must be a string, Pathlike or None')
            if isinstance(uri, Path):
                uri = uri.as_posix()

            if uri.startswith('http://') or uri.startswith('https://'):
                try:
                    response = httpx.get(uri)
                    if response.status_code != 200:
                        raise ConnectionError(f'Failed to connect to the server at {uri}.')

                    try:
                        rj = response.json()
                        if rj != {'status': 'success', 'message': 'LynseDB HTTP API'}:
                            raise ConnectionError(f'Failed to connect to the server at {uri}.')
                    except Exception as e:
                        print(e)
                        raise ConnectionError(f'Failed to connect to the server at {uri}.')

                except httpx.RequestError:
                    raise ConnectionError(f'Failed to connect to the server at {uri}.')
        else:
            uri = Path(uri) if uri is not None else None

        self._uri = uri

    def create_database(self, database_name: str, drop_if_exists: bool = False):
        """
        Create the database using a lazy mode, where entities are only created when they are actually used.

        Parameters:
            database_name (str): The name of the database to create.
            drop_if_exists (bool): Whether to drop the database if it already exists.
                If set to True, the existing database will be immediately deleted before creating a new one.

        Returns:
            None
        """
        from .api.http_api.client_api import raise_error_response
        from .configs.config import config
        import httpx

        if not self._is_remote:
            from .api.native_api.database_manager import DatabaseManager

            db_manager = DatabaseManager(root_path=self._uri or config.LYNSE_DEFAULT_ROOT_PATH)
            db_manager.register(db_name=database_name)

            if drop_if_exists:
                _InstanceDistributor(root_path=self._uri or config.LYNSE_DEFAULT_ROOT_PATH,
                                     database_name=database_name).drop_database()
        else:
            try:
                rj = httpx.post(f'{self._uri}/create_database', json={'database_name': database_name,
                                                                      'drop_if_exists': drop_if_exists})
                if rj.status_code != 200:
                    raise_error_response(rj)
            except httpx.RequestError:
                raise ConnectionError(f'Failed to connect to the server at {self._uri}.')

        return _InstanceDistributor(root_path=self._uri or config.LYNSE_DEFAULT_ROOT_PATH, database_name=database_name)

    def get_database(self, database_name: str):
        """
        Get an existing database.

        Parameters:
            database_name (str): The name of the database to get.

        Returns:
            LynseDB: (LocalClient, HTTPClient):
                The appropriate LynseDB client instance based on the root path.
                If the root path is a local path, return a LocalClient instance,
                otherwise return a HTTPClient instance.
        """
        from spinesUtils.asserts import raise_if

        from .api.http_api.client_api import raise_error_response
        from .configs.config import config
        import httpx

        if not self._is_remote:
            from .api.native_api.database_manager import DatabaseManager

            db_manager = DatabaseManager(root_path=self._uri or config.LYNSE_DEFAULT_ROOT_PATH)
            databases = db_manager.list_database()
            raise_if(ValueError, database_name not in databases, f'{database_name} does not exist.')
        else:
            try:
                rj = httpx.get(f'{self._uri}/list_databases')
                if rj.status_code != 200:
                    raise_error_response(rj)

                databases = rj.json()['params']['databases']
                raise_if(ValueError, database_name not in databases, f'{database_name} does not exist.')
            except httpx.RequestError:
                raise ConnectionError(f'Failed to connect to the server at {self._uri}.')

        return _InstanceDistributor(root_path=self._uri or config.LYNSE_DEFAULT_ROOT_PATH, database_name=database_name)

    def list_databases(self):
        """
        List all databases.

        Returns:
            List: A list of all databases.
        """
        from .api.http_api.client_api import raise_error_response
        import httpx

        if not self._is_remote:
            from .api.native_api.database_manager import DatabaseManager
            from .configs.config import config

            db_manager = DatabaseManager(root_path=self._uri or config.LYNSE_DEFAULT_ROOT_PATH)
            return db_manager.list_database()
        else:
            try:
                rj = httpx.get(f'{self._uri}/list_databases')
                if rj.status_code != 200:
                    raise_error_response(rj)

                return rj.json()['params']['databases']
            except httpx.RequestError:
                raise ConnectionError(f'Failed to connect to the server at {self._uri}.')

    def drop_database(self, database_name: str):
        """
        Delete a database.

        Parameters:
            database_name (str): The name of the database to delete.

        Returns:
            None
        """
        from .api.http_api.client_api import raise_error_response
        import httpx

        if not self._is_remote:
            from .api.native_api.database_manager import DatabaseManager
            from .configs.config import config

            db_manager = DatabaseManager(root_path=self._uri or config.LYNSE_DEFAULT_ROOT_PATH)
            databases = db_manager.list_database()
            if database_name not in databases:
                return
            db_manager.delete(database_name)
        else:
            try:
                rj = httpx.post(f'{self._uri}/delete_database', json={'database_name': database_name})
                if rj.status_code != 200:
                    raise_error_response(rj)
            except httpx.RequestError:
                raise ConnectionError(f'Failed to connect to the server at {self._uri}.')

    def __repr__(self):
        from .configs.config import config
        return f'{self.__class__.__name__}(uri={self._uri or "DefaultRootPath"})'

    def __str__(self):
        return self.__repr__()


def _load_and_register_module(module):
    """Dynamically import and register a module under the current module namespace."""
    import importlib
    import sys

    rename_modules = {'fields_cache': 'field_models'}

    try:
        if isinstance(module, str):
            module_name = module.rsplit('.', 1)[-1]
            module = importlib.import_module(module)
        else:
            module_name = module.__name__.rsplit('.', 1)[-1]

        module_name = rename_modules.get(module_name, module_name)

    except ImportError as e:
        raise ImportError(f"Could not import module {module}") from e

    current_module_name = __name__
    sys.modules[f'{current_module_name}.{module_name}'] = module


_load_and_register_module(field_models)



import json
from functools import wraps
from pathlib import Path

from spinesUtils.asserts import raise_if, generate_function_kwargs
from spinesUtils.logging import Logger


class ParametersValidator:
    """Database configuration, define once and can not be changed.

    """
    def __init__(self, immutable_param: list, logger: Logger):
        raise_if(TypeError, not isinstance(immutable_param, list), "update_configs must be a list.")

        self.update_configs_list = immutable_param
        self.logger = logger

    def check_configs(self, configs: dict, update_configs_dict: dict):
        """Check if the configurations are set."""
        for key, value in configs.items():
            if key not in update_configs_dict:
                continue
            else:
                if update_configs_dict[key] != value:
                    self.logger.warning(f"The immutable parameter `{key}` of the database has been set to {value}, "
                                        f"the new value {update_configs_dict[key]} will be ignored.")

        for key, value in update_configs_dict.items():
            if key not in configs:
                self.logger.warning(f"The database configuration {key} is not set, "
                                    f"and it will be set to {value}.")
                configs[key] = value

        return configs

    def load_configs(self, configs_json: Path):
        """Load the configurations."""
        try:
            with open(configs_json, 'r') as f:
                configs = json.load(f)
            return configs
        except (FileNotFoundError, json.JSONDecodeError, PermissionError, OSError) as e:
            self.logger.error(f"Failed to load the LynseDB configurations.")
            raise e

    def save_configs(self, configs_json: Path, configs: dict):
        """Save the configurations."""
        try:
            with open(configs_json, 'w') as f:
                json.dump(configs, f)

            return configs
        except (PermissionError, OSError) as e:
            self.logger.error(f"Failed to save the LynseDB configurations.")
            raise e

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # first parameter is self
            self_instance = args[0]

            kwargs = generate_function_kwargs(func, *args, **kwargs)

            dir_path = Path(kwargs.get("database_path"))

            update_configs_dict = {}
            for key in self.update_configs_list:
                update_configs_dict[key] = kwargs.get(key)

            self.database_path_parent = Path(dir_path).parent.absolute() / dir_path.absolute().name

            if not self.database_path_parent.exists():
                self.database_path_parent.mkdir(parents=True)
                first_create = True
            else:
                first_create = False

            # rename the configs file, the version upgrade has resulted in compatibility modifications
            if (self.database_path_parent / 'configs.json').exists():
                (self.database_path_parent / 'configs.json').rename(self.database_path_parent / 'pv-configs.json')

            configs_json = self.database_path_parent / Path('pv-configs.json')

            if first_create or not configs_json.exists():
                final_configs = self.save_configs(configs_json, update_configs_dict)
            else:
                existing_configs = self.load_configs(configs_json)
                final_configs = self.check_configs(existing_configs, update_configs_dict)

            for i in kwargs:
                if i not in final_configs and kwargs[i] != self_instance:
                    final_configs[i] = kwargs[i]

            return func(self_instance, **final_configs)

        return wrapper



from abc import ABC, abstractmethod

from spinesUtils.asserts import raise_if

from ..core_components.fast_sort import FastSort


class BaseIndex(ABC):
    """Defines the interface for an index."""

    def __init__(self):
        self.distance = None

    def _register_distance(self, distance):
        """Registers the distance function."""
        self.distance = distance

    @abstractmethod
    def encode(self, *args, **kwargs):
        """Encodes the data."""
        pass

    @abstractmethod
    def fit_transform(self, *args, **kwargs):
        """Fits the model and transforms the data."""
        pass

    def check_and_encode(self, original_vec=None, encoded_vec=None, original_data=None,
                         encoded_data=None):
        """Searches for nearest neighbors."""
        raise_if(ValueError, original_data is None and encoded_data is None,
                 "Either original_data or encoded_data must be provided.")
        raise_if(ValueError, original_vec is None and encoded_vec is None,
                 "Either original_vec or encoded_vec must be provided.")

        if encoded_data is None:
            encoded_data = self.encode(original_data)
        if encoded_vec is None:
            encoded_vec = self.encode(original_vec)

        return encoded_vec, encoded_data

    @abstractmethod
    def save(self, filepath):
        """Saves the model to a file."""
        pass

    @abstractmethod
    def load(self, filepath):
        """Loads the model from a file."""
        pass



from concurrent.futures import ThreadPoolExecutor

import numpy as np
from lynse.computational_layer.engines import inner_product
from spinesUtils.asserts import check_has_param

from ..configs.config import config
from ..core_components.fields_cache import ExpressionParser
from ..execution_layer.matrix_serializer import MatrixSerializer
from ..utils.utils import SearchResultsCache
from ..core_components.limited_sort import LimitedSorted


class Search:
    """Search the database for the vectors most similar to the given vector."""

    def __init__(self,
                 matrix_serializer: 'MatrixSerializer',
                 n_threads=10) -> None:
        """
        Search the database for the vectors most similar to the given vector.

        Parameters:
            matrix_serializer (MatrixSerializer): The database to be queried.
            n_threads (int): The number of threads to use for searching the database.
        """
        self.matrix_serializer = matrix_serializer
        self.logger = self.matrix_serializer.logger
        # attributes
        self.dtypes = self.matrix_serializer.dtypes
        self.chunk_size = self.matrix_serializer.chunk_size

        self.fields_index = self.matrix_serializer.field_index

        self.n_threads = n_threads

        self.executors = ThreadPoolExecutor(max_workers=self.n_threads)

    def _narrow_down_ids(self, search_filter):
        """
        Narrow down the search results by applying the filter.

        Parameters:
            search_filter (Filter): The filter to apply to the search.

        Returns:
            np.ndarray: The narrowed down indices.
        """
        return None if search_filter is None else self.matrix_serializer.field_index.query(search_filter)

    def _search_chunk(self, vector, subset_indices, filename, limited_sorted, distance_func, ivf_subset_indices=None,
                      topk=10):
        """
        Search a single database chunk for the vectors most similar to the given vector.

        Parameters:
            vector (np.ndarray): The search vector.
            subset_indices (np.ndarray): The indices to filter the numpy array.
            filename (str): The name of the database chunk to search.
            limited_sorted (LimitedSorted): The object to store the nearest vectors.
            distance_func (function): The distance function to use for the search.
            ivf_subset_indices (np.ndarray): The indices to filter the numpy array in the IVF index.
            topk (int): The number of nearest vectors to return.

        Returns:
            Tuple: The indices and similarity scores of the nearest vectors in the chunk.
        """
        if ivf_subset_indices is not None:
            if subset_indices is None:
                subset_indices = ivf_subset_indices
            else:
                subset_indices = np.intersect1d(subset_indices, ivf_subset_indices)

        cache = self.matrix_serializer.storage_worker.dataloader.cache
        if subset_indices is not None:
            if cache is None or filename not in cache:
                database_chunk, index_chunk = self.matrix_serializer.storage_worker.mmap_read(filename)
                filter_indices = np.isin(index_chunk, subset_indices)
                database_chunk = database_chunk[filter_indices]
                index_chunk = index_chunk[filter_indices]
            else:
                database_chunk, index_chunk = self.matrix_serializer.storage_worker.dataloader.cache[filename]
                filter_indices = np.isin(index_chunk, subset_indices)
                database_chunk = database_chunk[filter_indices]
                index_chunk = index_chunk[filter_indices]
        else:
            if cache is None or filename not in cache:
                database_chunk, index_chunk = self.matrix_serializer.storage_worker.mmap_read(filename)
            else:
                database_chunk, index_chunk = self.matrix_serializer.storage_worker.dataloader.cache[filename]

        if len(index_chunk) == 0:
            return [], [], []

        # Distance calculation core code
        topk = min(topk, len(index_chunk))  # make sure topk is not larger than the chunk size
        res = distance_func(original_vec=vector, encoded_data=database_chunk, top_k=topk)

        if len(res) == 2 and isinstance(res[0], np.ndarray) and isinstance(res[1], np.ndarray):
            ids, scores = res
            index_chunk = index_chunk[ids]
        else:
            scores = res, None

        limited_sorted.add(scores, index_chunk)

    def _flat_search(self, vector, k, search_filter):
        """
        Search the database for the vectors most similar to the given vector in FLAT mode.

        Parameters:
            vector (np.ndarray): The search vector.
            k (int): The number of nearest vectors to return.
            search_filter (Filter): The field filter to apply to the search.

        Returns:
            Tuple: If return_fields is True, the indices, similarity scores,
                    and fields of the nearest vectors in the database.
                Otherwise, the indices and similarity scores of the nearest vectors in the database.
        """
        vector = vector.astype(self.dtypes) if vector.dtype != self.dtypes else vector

        subset_indices = self._narrow_down_ids(search_filter)

        npy_filenames = self.matrix_serializer.storage_worker.get_all_files(separate=False)

        limited_sorted = LimitedSorted(n=k)

        def batch_search(is_ivf=True, cid=None, directly_return=True):
            nonlocal subset_indices, npy_filenames

            ivf_subset_indices = None if not is_ivf else self.matrix_serializer.indexer.ivf.ivf_index.get_entries(cid)

            distance_func = self.matrix_serializer.indexer.index.search

            if is_ivf:
                npy_filenames = list(filter(lambda x: x in ivf_subset_indices, npy_filenames))

            if npy_filenames:
                _ = [
                    i for i in map(
                        lambda x: self._search_chunk(
                            vector, subset_indices, x,
                            limited_sorted,
                            distance_func,
                            ivf_subset_indices[x] if ivf_subset_indices is not None else None,
                            topk=k
                        ), npy_filenames
                    )
                ]

            if directly_return:
                res_ids, res_scores = limited_sorted.get_top_n()

                return (
                    res_ids, res_scores if self.matrix_serializer.indexer.index_mode != 'Flat-IP' else -1 * res_scores
                )

        # if the index mode is FLAT, use FLAT
        if self.matrix_serializer.indexer.index_mode.startswith('Flat'):
            return batch_search(is_ivf=False)

        # otherwise, use IVF-FLAT
        cluster_id_sorted = inner_product(
            vector, self.matrix_serializer.indexer.ivf.ann_model.cluster_centers_,
            n=self.matrix_serializer.indexer.ivf.ann_model.cluster_centers_.shape[0],
            use_simd=False
        )[0].tolist()

        for cluster_id in cluster_id_sorted:
            batch_search(cid=cluster_id, directly_return=False, is_ivf=True)

            if len(limited_sorted) >= k:
                break

        res_ids, res_scores = limited_sorted.get_top_n()
        return (
            res_ids, res_scores if self.matrix_serializer.indexer.index_mode != 'IVF-IP' else -1 * res_scores
        )

    @SearchResultsCache(config.LYNSE_SEARCH_CACHE_SIZE, config.LYNSE_SEARCH_CACHE_EXPIRE_SECONDS)
    def single_search(self, vector, k=12, search_filter=None, return_fields=False, **kwargs):
        """
        Search the database for the vectors most similar to the given vector in batches.
        """
        if self.matrix_serializer.indexer.index_mode.split('-')[-1] not in ['SQ8', 'Binary']:
            res_ids, res_scores = self._flat_search(vector, k, search_filter)
        else:
            params = {
                'original_vec': vector,
                'top_k': k,
                'subset_indices': self._narrow_down_ids(search_filter)
            }

            if check_has_param(self.matrix_serializer.indexer.index.search, 'rescore_multiplier'):
                if 'Binary' in self.matrix_serializer.indexer.index_mode:
                    params['rescore_multiplier'] = kwargs.get('rescore_multiplier', 10)
                else:
                    params['rescore_multiplier'] = kwargs.get('rescore_multiplier', 2)

            if check_has_param(self.matrix_serializer.indexer.index.search, 'rescore'):
                params['rescore'] = kwargs.get('rescore', False)

            res_ids, res_scores = self.matrix_serializer.indexer.index.search(**params)

        if return_fields:
            res_fields = self.fields_index.retrieve_ids(res_ids.tolist(), include_external_id=True)
        else:
            res_fields = None

        return res_ids, res_scores, res_fields

    def multi_search(self, vectors, k=12, search_filter=None, return_fields=False, **kwargs):
        """
        Search the database for the vectors most similar to the given vectors in batches.
        """
        ids = []
        scores = []
        fields = []

        for i, vector in enumerate(vectors):
            res_ids, res_scores, res_fields = self.single_search(vector, k, search_filter, return_fields, **kwargs)
            ids.append(res_ids)
            scores.append(res_scores)
            fields.append(res_fields)

        return np.vstack(ids), np.vstack(scores), fields

    def search(self, vector, k=12, search_filter=None, return_fields=False, **kwargs):
        """
        Search the database for the vectors most similar to the given vector.

        Parameters:
            vector (np.ndarray or list): The search vectors, it can be a single vector or a list of vectors.
                The vectors must have the same dimension as the vectors in the database,
                and the type of vector can be a list or a numpy array.
            k (int): The number of nearest vectors to return.
            search_filter (Filter or FilterExpression string, optional): The filter to apply to the search.
            return_fields (bool): Whether to return the fields of the search results.
            kwargs: Additional keyword arguments. The following are valid:
                rescore (bool): Whether to rescore the results of binary or scaler quantization searches.
                    Default is False. It is recommended to set it to True when the index mode is 'Binary'.
                rescore_multiplier (int): The multiplier for the rescore operation.
                    It is only available when rescore is True.
                    If 'Binary' is in the index mode, the default is 10. Otherwise, the default is 2.

        Returns:
            Tuple: If return_fields is True, the indices, similarity scores,
                    and fields of the nearest vectors in the database.
                Otherwise, the indices and similarity scores of the nearest vectors in the database.

        Raises:
            ValueError: If the database is empty.
        """
        if not hasattr(self.matrix_serializer, "indexer"):
            raise ValueError('The database has not been indexed.')

        if not isinstance(vector, np.ndarray):
            vector = np.asarray(vector).squeeze()

        if isinstance(search_filter, str):
            search_filter = ExpressionParser(search_filter).to_filter()

        if vector.shape[0] > 1:
            return self.multi_search(vector, k, search_filter, return_fields, **kwargs)
        else:
            return self.single_search(vector, k, search_filter, return_fields, **kwargs)

    def __del__(self):
        if self.executors is not None:
            self.executors.shutdown(wait=True)
            self.logger.debug('Search executor shutdown.')

    def delete(self):
        self.__del__()



import numpy as np
import json
import os
import shutil
from pathlib import Path

from spinesUtils.timer import Timer

from ..core_components.locks import ThreadLock


class WALStorage:
    def __init__(self, collection_name, chunk_size, storage_path, flush_interval=5):
        self.storage_path = Path(storage_path) / "wal"
        self.storage_path.mkdir(parents=True, exist_ok=True)

        self.log_dir = self.storage_path / f"log"
        self.read_dir = self.storage_path / f"read"
        self.state_dir = self.storage_path / f"state"

        # rename the folder, the version upgrade has resulted in compatibility modifications
        for _ in [self.storage_path / f"{collection_name}-log", self.storage_path / f"{collection_name}-read",
                  self.storage_path / f"{collection_name}-state"]:
            if _.exists():
                _.rename(self.storage_path / f"{_.name.split('-')[-1].strip()}")

        self.log_dir.mkdir(exist_ok=True)
        self.read_dir.mkdir(exist_ok=True)
        self.state_dir.mkdir(exist_ok=True)

        self.chunk_size = chunk_size
        self.flush_interval = flush_interval
        self.file_id = 0  # File ID to maintain the naming order
        self.lock = ThreadLock()

        # Buffer to store data, indices, and fields
        self.buffer_data = []
        self.buffer_indices = []
        self.buffer_fields = []

        self.timer = Timer()
        self.timer.start()

        # Start periodic flush thread
        self.running = True

    def write_log_data(self, data, indices, fields):
        with self.lock:
            if not isinstance(data, np.ndarray):
                data = np.vstack(data)

            if not isinstance(indices, np.ndarray):
                indices = np.array(indices)

            if not isinstance(fields, list) or not all(isinstance(field, dict) for field in fields):
                raise ValueError("fields should be a list of dictionaries")

            self.buffer_data.append(data)
            self.buffer_indices.append(indices)
            self.buffer_fields.extend(fields)

            # Check if buffer is full
            if sum(len(chunk) for chunk in self.buffer_data) >= self.chunk_size:
                self._flush_buffer_to_disk()

            if self.timer.last_timestamp_diff() >= self.flush_interval:
                self._flush_buffer_to_disk()
                self.timer.middle_point()

    def _flush_buffer_to_disk(self):
        with self.lock:
            # Check if buffer is empty
            if not self.buffer_data:
                return

            # Concatenate buffered data
            try:
                data = np.concatenate(self.buffer_data, axis=0)
                indices = np.concatenate(self.buffer_indices, axis=0)
                fields = self.buffer_fields
            except Exception as e:
                print(f"Error during concatenation: {e}")
                return

            num_rows = data.shape[0]
            start = 0

            while start < num_rows:
                end = min(start + self.chunk_size, num_rows)
                chunk_data = data[start:end]
                chunk_indices = indices[start:end]
                chunk_fields = fields[start:end]

                # Serialize fields using JSON
                chunk_fields_serialized = [json.dumps(field) for field in chunk_fields]

                # Save to NPZ file
                log_file_path = self.log_dir / f"log_{self.file_id}.npz"
                try:
                    np.savez(log_file_path, data=chunk_data, indices=chunk_indices,
                             fields=np.array(chunk_fields_serialized))
                except Exception as e:
                    print(f"Error during NPZ write: {e}")
                    return

                self.file_id += 1
                self._write_state_file(self.file_id - 1)

                start += self.chunk_size

            # Clear the buffer
            self.buffer_data = []
            self.buffer_indices = []
            self.buffer_fields = []

    def _write_state_file(self, file_id):
        state_file_path = self.state_dir / f"state_{file_id}.txt"
        try:
            with open(state_file_path, 'w') as f:
                f.write("COMPLETED")
        except Exception as e:
            print(f"Error during state file write: {e}")

    def _remove_state_file(self, file_id):
        state_file_path = self.state_dir / f"state_{file_id}.txt"
        if state_file_path.exists():
            os.remove(state_file_path)

    @staticmethod
    def read_log_data(log_filepath):
        try:
            with np.load(log_filepath, allow_pickle=True) as npzfile:
                data = npzfile['data']
                indices = npzfile['indices']
                fields_serialized = npzfile['fields']

                # Deserialize fields using JSON
                fields = []
                for i, field in enumerate(fields_serialized):
                    try:
                        fields.append(json.loads(field))
                    except Exception as e:
                        print(f"Error unpacking field at index {i}: {field}. Error: {e}")
                        raise

            return data, indices, fields
        except (IOError, OSError) as e:
            print(f"Error reading file {log_filepath}, it may be corrupted. Error: {e}")
            return None, None, None

    def get_file_iterator(self):
        with self.lock:
            self._flush_buffer_to_disk()

            file_ids = sorted([int(i.stem.split('_')[-1]) for i in self.log_dir.glob('log_*.npz')])
            for file_id in file_ids:
                log_path = self.log_dir / f"log_{file_id}.npz"
                state_file_path = self.state_dir / f"state_{file_id}.txt"
                if state_file_path.exists():
                    read_flag = self.read_dir / f"read_{file_id}.flag"
                    if not read_flag.exists():
                        data, indices, fields = self.read_log_data(log_path)
                        if data is not None and indices is not None and fields is not None:
                            yield data, indices, fields
                            read_flag.touch()  # Mark this file as read
                        else:
                            print(f"Skipping corrupted file: {log_path}")
                    self._remove_state_file(file_id)

    @property
    def file_number(self):
        return len(list(self.log_dir.glob('log_*.npz')))

    def cleanup(self):
        with self.lock:
            try:
                if self.storage_path.exists():
                    shutil.rmtree(self.storage_path)
            except Exception as e:
                print(f"Error during cleanup: {e}")

    def reincarnate(self):
        with self.lock:
            self.cleanup()
            self.storage_path.mkdir(parents=True, exist_ok=True)
            self.log_dir.mkdir(exist_ok=True)
            self.read_dir.mkdir(exist_ok=True)
            self.state_dir.mkdir(exist_ok=True)
            self.file_id = 0

    def stop(self):
        self.running = False



import os
from pathlib import Path
from typing import Union, List, Tuple

import numpy as np
from spinesUtils.asserts import raise_if, ParameterTypeAssert
from spinesUtils.timer import Timer

from ...configs.parameters_validator import ParametersValidator
from ...core_components.fields_cache import IndexSchema
from ...core_components.locks import ThreadLock
from ...execution_layer.indexer import Indexer
from ...execution_layer.search import Search
from ...execution_layer.matrix_serializer import MatrixSerializer
from ...utils.utils import unavailable_if_deleted, unavailable_if_empty, collection_repr
from ...api import logger
from ...core_components.fields_cache.filter import Filter


class ExclusiveDB:
    """
    A class for managing a vector database stored in chunk files and computing vectors similarity.
    The class is exclusive and cannot be shared with other threads or processes,
    so it is not thread-safe or process-safe.

    Note:
        This class is not called at the top level, but is called through the LocalClient class.
        When directly operating on data by calling this class, users need to be clear about what they are doing.
    """
    name = "Local"

    @ParametersValidator(
        immutable_param=['dim', 'chunk_size', 'dtypes'],
        logger=logger
    )
    @ParameterTypeAssert({
        'dim': int,
        'database_path': str,
        'chunk_size': int,
        'dtypes': str,
        'use_cache': bool,
        'n_threads': (None, int),
        'warm_up': bool,
        'cache_chunks': int
    }, func_name='ExclusiveDB')
    def __init__(
            self,
            dim: int,
            database_path: Union[str, Path],
            chunk_size: int = 100_000,
            dtypes: str = 'float32',
            use_cache: bool = True,
            n_threads: Union[int, None] = 10,
            warm_up: bool = False,
            cache_chunks: int = 20
    ):
        """
        Initialize the vector database.

        Parameters:
            dim (int): Dimension of the vectors.
            database_path (str or Path): The path to the database file.
            chunk_size (int): The size of each data chunk. Default is 100,000.
                It's recommended to be between 10,000 and 500,000.
            dtypes (str): The data type of the vectors. Default is 'float32'.
                Options are 'float16', 'float32' or 'float64'.
            use_cache (bool): Whether to use cache for search. Default is True.
            n_threads (int): The number of threads to use for parallel processing. Default is 10.
            warm_up (bool): Whether to warm up the database. Default is False.
            cache_chunks (int): The number of chunks to cache in memory. Default is 20.

        Raises:
            ValueError: If `chunk_size` is less than or equal to 1.
        """
        raise_if(ValueError, chunk_size <= 1, 'chunk_size must greater than 1')
        raise_if(ValueError, dtypes not in ('float16', 'float32', 'float64'),
                 'dtypes must be "float16", "float32" or "float64')

        if chunk_size <= 1:
            raise ValueError('chunk_size must be greater than 1')

        if not 10000 <= chunk_size <= 500000:
            logger.warning('The recommended chunk size is between 10,000 and 500,000.')

        # In order to have enough data to index the segment
        raise_if(ValueError, chunk_size < 1000, 'chunk_size must greater than or equal to 1000.')

        self._database_path = database_path
        self._database_name = Path(database_path).parent.name
        self._collection_name = Path(database_path).name

        self._matrix_serializer = MatrixSerializer(
            dim=dim,
            collection_path=self._database_path,
            chunk_size=chunk_size,
            logger=logger,
            dtypes=dtypes,
            warm_up=warm_up,
            cache_chunks=cache_chunks
        )
        self._data_loader = self._matrix_serializer.dataloader
        self._id_filter = self._matrix_serializer.id_filter

        self._timer = Timer()
        self._use_cache = use_cache

        raise_if(TypeError, n_threads is not None and not isinstance(n_threads, int), "n_threads must be an integer.")
        raise_if(ValueError, n_threads is not None and n_threads <= 0, "n_threads must be greater than 0.")

        self._indexer = Indexer(
            logger=logger,
            dataloader=self._data_loader,
            storage_worker=self._matrix_serializer.storage_worker,
            collections_path_parent=self._matrix_serializer.collections_path_parent
        )

        self._search = Search(
            matrix_serializer=self._matrix_serializer,
            n_threads=n_threads if n_threads else min(32, os.cpu_count() + 4),
        )

        self._search.single_search.clear_cache()

        # initial lock
        self._lock = ThreadLock()

        # pre_build_index if the ExclusiveDB instance is reloaded
        if self.shape[0] > 0:
            self._pre_build_index()

    @unavailable_if_deleted
    def add_item(self, vector: Union[np.ndarray, list], id: int, *, field: dict = None,
                 buffer_size: Union[int, None, bool] = True):
        """
        Add a single vector to the collection.

        It is recommended to use incremental ids for best performance.

        Parameters:
            vector (np.ndarray): The vector to be added.
            id (int): The ID of the vector.
            field (dict, optional, keyword-only): The field of the vector. Default is None.
                If None, the field will be set to an empty string.
            buffer_size (int or bool or None): The buffer size for the storage worker. Default is True.

                - If None, the vector will be directly written to the disk.
                - If True, the buffer_size will be set to chunk_size,
                    and the vectors will be written to the disk when the buffer is full.
                - If False, the vector will be directly written to the disk.
                - If int, when the buffer is full, the vectors will be written to the disk.

        Returns:
            (int): The ID of the added vector.

        Raises:
            ValueError: If the vector dimensions don't match or the ID already exists.
        """
        return self._matrix_serializer.add_item(vector, id=id, field=field,
                                                buffer_size=buffer_size)

    @unavailable_if_deleted
    def bulk_add_items(
            self, vectors: Union[List[Tuple[np.ndarray, int, dict]], List[Tuple[np.ndarray, int]]],
            **kwargs
    ):
        """
        Bulk add vectors to the collection in batches.

        It is recommended to use incremental ids for best performance.

        Parameters:
            vectors (list or tuple): A list or tuple of vectors to be saved.
                Each vector can be a tuple of (vector, id, field).

        Returns:
            List[int]: A list of indices where the vectors are stored.
        """
        return self._matrix_serializer.bulk_add_items(vectors)

    @unavailable_if_deleted
    def commit(self):
        """
        Save the database, ensuring that all data is written to disk.
        """
        with self._lock:
            self._matrix_serializer.commit()
            self._pre_build_index()

    @unavailable_if_deleted
    def _pre_build_index(self):
        if not hasattr(self._matrix_serializer, "indexer"):
            logger.info("Pre-building the index...")
            self.build_index(index_mode="Flat-IP")  # Default index mode

    @unavailable_if_deleted
    @unavailable_if_empty
    def build_index(self, index_mode, rebuild=False, **kwargs):
        """
        Build the index for clustering.

        Parameters:
            index_mode (str): The index mode, must be one of the following:

                - 'IVF-IP-SQ8': IVF index with inner product and scalar quantizer with 8 bits.
                    The distance is inner product.
                - 'IVF-IP': IVF index with inner product. (Alias: 'IVF')
                - 'IVF-L2sq-SQ8': IVF index with squared L2 distance and scalar quantizer with 8 bits.
                    The distance is squared L2 distance.
                - 'IVF-L2sq': IVF index with squared L2 distance.
                - 'IVF-Cos-SQ8': IVF index with cosine similarity and scalar quantizer with 8 bits.
                    The distance is cosine similarity.
                - 'IVF-Cos': IVF index with cosine similarity.
                - 'IVF-Jaccard-Binary': IVF index with binary quantizer. The distance is Jaccard distance.
                - 'IVF-Hamming-Binary': IVF index with binary quantizer. The distance is Hamming distance.
                - 'Flat-IP-SQ8': Flat index with inner product and scalar quantizer with 8 bits.
                - 'Flat-IP': Flat index with inner product. (Alias: 'FLAT')
                - 'Flat-L2sq-SQ8': Flat index with squared L2 distance and scalar quantizer with 8 bits.
                - 'Flat-L2sq': Flat index with squared L2 distance.
                - 'Flat-Cos-SQ8': Flat index with cosine similarity and scalar quantizer with 8 bits.
                - 'Flat-Cos': Flat index with cosine similarity.
                - 'Flat-Jaccard-Binary': Flat index with binary quantizer. The distance is Jaccard distance.
                - 'Flat-Hamming-Binary': Flat index with binary quantizer. The distance is Hamming distance.
            rebuild (bool): Whether to rebuild the index.
            kwargs: Additional keyword arguments. The following are available:

                - 'n_clusters' (int): The number of clusters. It is only available when the index_mode including 'IVF'.

        Returns:
            None
        """
        with self._lock:
            self._indexer.build_index(index_mode, rebuild=rebuild, **kwargs)
            self._matrix_serializer.indexer = self._indexer

    @unavailable_if_deleted
    @unavailable_if_empty
    def remove_index(self):
        """
        Remove the vector index.

        Returns:
            None
        """
        self._indexer.remove_index()
        logger.info("Fallback to default index mode: `Flat-IP`.")
        self.build_index(index_mode='Flat-IP')  # Default index mode

    @unavailable_if_deleted
    @unavailable_if_empty
    def search(self, vector: Union[np.ndarray, list], k: int = 10, *,
               search_filter: Filter = None, return_fields=False, **kwargs):
        """
        Search the database for the vectors most similar to the given vector.

        Parameters:
            vector (np.ndarray or list): The search vectors, it can be a single vector or a list of vectors.
                The vectors must have the same dimension as the vectors in the database,
                and the type of vector can be a list or a numpy array.
            k (int): The number of nearest vectors to return.
            search_filter (Filter or FilterExpression string, optional): The filter to apply to the search.
            return_fields (bool): Whether to return the fields of the search results.
            kwargs: Additional keyword arguments. The following are valid:

                - rescore (bool): Whether to rescore the results of binary or scaler quantization searches.
                    Default is False. It is recommended to set it to True when the index mode is 'Binary'.
                - rescore_multiplier (int): The multiplier for the rescore operation.
                    It is only available when rescore is True.
                    If 'Binary' is in the index mode, the default is 10. Otherwise, the default is 2.

        Returns:
            (Tuple[List[int], List[float], List[Dict]] or Tuple[List[int], List[float], None]):
                If return_fields is True, the indices, similarity scores,
                and fields of the nearest vectors in the database.
                Otherwise, the indices and similarity scores of the nearest vectors in the database.

        Raises:
            ValueError: If the database is empty.
        """
        raise_if(ValueError, not isinstance(vector, (np.ndarray, list)),
                 'vector must be np.ndarray or list.')

        raise_if(TypeError, not isinstance(k, int) and not (isinstance(k, str) and k != 'all'),
                 'k must be int or "all".')
        raise_if(ValueError, k <= 0, 'k must be greater than 0.')
        raise_if(ValueError, not isinstance(search_filter, (Filter, type(None), str)),
                 'search_filter must be Filter or None or FieldExpression string.')

        # Convert the vector to np.ndarray
        vector = np.atleast_2d(np.asarray(vector))

        raise_if(ValueError, vector.shape[1] != self._matrix_serializer.shape[1],
                 'vector must be same dim with database.')

        rescore = kwargs.setdefault('rescore', False)
        rescore_multiplier = kwargs.setdefault('rescore_multiplier', 2)
        raise_if(TypeError, not isinstance(rescore, bool), 'rescore must be bool.')
        raise_if(TypeError, not isinstance(rescore_multiplier, int), 'rescore_multiplier must be int.')

        if self._matrix_serializer.shape[0] == 0:
            raise ValueError('database is empty.')

        if k > self._matrix_serializer.shape[0]:
            k = self._matrix_serializer.shape[0]

        if not self._use_cache:
            self._search.single_search.clear_cache()

        res = self._search.search(vector=vector, k=k, search_filter=search_filter,
                                  return_fields=return_fields, **kwargs)

        return res

    @unavailable_if_deleted
    def is_id_exists(self, id):
        """
        Check if the ID exists in the database.

        Parameters:
            id (int): The ID to check.

        Returns:
            Bool: True if the ID exists, False otherwise.
        """
        return id in self._matrix_serializer.id_filter

    @property
    @unavailable_if_deleted
    @unavailable_if_empty
    def max_id(self):
        """
        Return the maximum ID in the database.

        Returns:
            (int): The maximum ID in the database.
        """
        return self._matrix_serializer.id_filter.find_max_value()

    @unavailable_if_deleted
    @unavailable_if_empty
    def query(self, query_filter, filter_ids=None, return_ids_only=False):
        """
        Query the fields cache.

        Parameters:
            query_filter (str): Filter or dict or FieldExpression string
                The filter object or the specify data to filter.
            filter_ids (List[int]):
                The list of external IDs to filter.
            return_ids_only (bool):
                If True, only the external IDs will be returned.

        Returns:
            (List[Dict]): The records. If not return_ids_only, the records will be returned.
            (List[int]): The external IDs. If return_ids_only, the external IDs will be returned.
        """
        return self._matrix_serializer.field_index.query(query_filter, filter_ids, return_ids_only)

    @unavailable_if_deleted
    @unavailable_if_empty
    def query_vectors(self, query_filter, filter_ids=None):
        """
        Query the vector data by the filter.

        Parameters:
            query_filter (Filter or dict or FieldExpression str or None):
                The filter object or the specify data to filter.
            filter_ids (list[int]):
                The list of external IDs to filter. Default is None.

        Returns:
            (Tuple[List[np.ndarray], List[int], List[Dict]]): The vectors, IDs, and fields of the items.
        """
        ids = self._matrix_serializer.field_index.query(query_filter, filter_ids, return_ids_only=True)
        if not ids:
            return np.array([]), np.array([]), []

        return self.read_by_only_id(ids)

    @property
    def shape(self):
        """
        Return the shape of the entire database.

        Returns:
            (Tuple[int, int]): The number of vectors and the dimension of each vector in the database.
        """
        return self._matrix_serializer.shape

    @unavailable_if_deleted
    def insert_session(self):
        """
        Create a session to insert data, which will automatically commit the data when the session ends.

        Returns:
            DataOpsSession (DataOpsSession): The session object.
        """
        from ...execution_layer.session import DataOpsSession

        return DataOpsSession(self)

    @unavailable_if_deleted
    @unavailable_if_empty
    def head(self, n=5):
        """
        Return the first n vectors in the database.

        Parameters:
            n (int): The number of vectors to return. Default is 5.

        Returns:
            (Tuple[List[np.ndarray], List[int], List[Dict]]): The vectors, IDs, and fields of the items.
        """
        filenames = self._matrix_serializer.storage_worker.get_all_files()
        filenames = sorted(filenames, key=lambda x: int(x.split('_')[-1].split('.')[0]))

        data, indices = [], []

        count = 0
        dtypes = self._matrix_serializer.dtypes

        for filename in filenames:
            data_chunk, index_chunk = self._matrix_serializer.dataloader(filename)
            dtypes = data_chunk.dtype
            data.extend(data_chunk)
            indices.extend(index_chunk)
            count += len(index_chunk)
            if count >= n:
                break

        if data:
            return (np.asarray(indices)[:n], np.vstack(data, dtype=dtypes)[:n],
                    self._matrix_serializer.field_index.retrieve_ids(indices[:n],
                                                                     include_external_id=True))
        return np.asarray(indices), np.array(data, dtype=dtypes), []

    @unavailable_if_deleted
    @unavailable_if_empty
    def tail(self, n=5):
        """
        Return the last n vectors in the database.

        Parameters:
            n (int): The number of vectors to return. Default is 5.

        Returns:
            (Tuple[List[np.ndarray], List[int], List[Dict]]): The vectors, IDs, and fields of the items.
        """
        filenames = self._matrix_serializer.storage_worker.get_all_files()
        filenames = sorted(filenames, key=lambda x: int(x.split('_')[-1].split('.')[0]))

        data, indices = [], []

        count = 0
        dtypes = self._matrix_serializer.dtypes

        for filename in filenames[::-1]:
            data_chunk, index_chunk = self._matrix_serializer.dataloader(filename)
            dtypes = data_chunk.dtype
            data.insert(0, data_chunk)
            indices = index_chunk.tolist() + indices
            count += len(index_chunk)
            if count >= n:
                break

        if data:
            return (np.asarray(indices)[-n:], np.vstack(data, dtype=dtypes)[-n:],
                    self._matrix_serializer.field_index.retrieve_ids(indices[-n:],
                                                                     include_external_id=True))
        return np.asarray(indices), np.array(data, dtype=dtypes), []

    @unavailable_if_deleted
    @unavailable_if_empty
    def yield_every_single_element(self, limit=None):
        """
        Yield every single element in the database.

        Parameters:
            limit (int): The maximum number of elements to yield. Default is None.

        Returns:
            (Tuple[np.ndarray, int, Dict]): The vectors, IDs, and fields of the items

        """
        raise_if(TypeError, limit is not None and not isinstance(limit, int), "limit must be an integer.")
        filenames = self._matrix_serializer.storage_worker.get_all_files()
        filenames = sorted(filenames, key=lambda x: int(x.split('_')[-1].split('.')[0]))

        count = 0
        for filename in filenames:
            data_chunk, index_chunk = self._matrix_serializer.dataloader(filename)
            for i in range(len(index_chunk)):
                yield int(index_chunk[i]), np.asarray(data_chunk[i]), \
                    self._matrix_serializer.field_index.retrieve([index_chunk[i]], include_external_id=True)
                count += 1
                if limit and count >= limit:
                    return

    @unavailable_if_deleted
    @unavailable_if_empty
    def read_by_only_id(self, id: Union[int, list]):
        """
        Read the vector data by the external ID.

        Parameters:
            id (int or list): The external ID or list of external IDs.

        Returns:
            (Tuple[np.ndarray, List[int], List[Dict]], Tuple[np.ndarray, int, Dict]):
                The vectors, IDs, and fields of the items.
        """
        data, ids = self._matrix_serializer.storage_worker.read_by_only_id(id)

        if data.shape[0] > 0:
            return np.asarray(ids), np.asarray(data), \
                self._matrix_serializer.field_index.retrieve_ids(ids, include_external_id=True)
        return np.asarray(ids), np.asarray(data), []

    @unavailable_if_deleted
    @unavailable_if_empty
    def list_fields(self):
        """
        Return all field names and their types.

        Returns:
            Dict: A dictionary with field names as keys and field types as values.
        """
        return self._matrix_serializer.field_index.storage.list_fields()

    @unavailable_if_deleted
    @unavailable_if_empty
    def build_field_index(self, schema, rebuild_if_exists=False):
        """
        Build an index for the field.

        Parameters:
            schema (IndexSchema or Field name string): The schema of the field or the field name string.
                When passing the field name string, the field name must be wrapped with ':', like ':vector:', ':timestamp:'.
            rebuild_if_exists (bool): Whether to rebuild the index if it already exists.

        Returns:
            None

        Note:
            The :id: is a reserved field name and cannot be used.
        """
        if not isinstance(schema, (IndexSchema, str)):
            raise TypeError("schema must be an instance of IndexSchema or a field string.")
        self._matrix_serializer.field_index.build_index(schema, rebuild_if_exists=rebuild_if_exists)

    @unavailable_if_deleted
    @unavailable_if_empty
    def remove_field_index(self, field_name):
        """
        Remove the index for the field.

        Parameters:
            field_name (str): The name of the field.

        Returns:
            None
        """
        if not field_name.startswith(':') or not field_name.endswith(':'):
            raise ValueError("The field name must be wrapped with ':'.")

        field_name = field_name.strip(':')
        self._matrix_serializer.field_index.remove_index(field_name)

    @unavailable_if_deleted
    @unavailable_if_empty
    def remove_all_field_indices(self):
        """
        Remove all the field indices.

        Returns:
            None
        """
        self._matrix_serializer.field_index.remove_all_field_indices()

    @unavailable_if_deleted
    @unavailable_if_empty
    def list_field_index(self):
        """
        List the field index.

        Returns:
            List: The list of field indices.
        """
        return self._matrix_serializer.field_index.list_indices()

    def delete(self):
        """
        Delete the database.

        This API is not part of the Collection in the HTTP API, but rather part of the HTTPClient in the HTTP API
        Users must call this API through the HTTPClient when they want to delete a collection on the server.

        Returns:
            None
        """
        if self._matrix_serializer.IS_DELETED:
            return

        import gc

        with self._lock:
            self._search.single_search.clear_cache()
            self._search.delete()
            self._matrix_serializer.delete()

        gc.collect()

    @property
    @unavailable_if_empty
    def index_mode(self):
        """
        Return the index mode of the database.

        Returns:
            (str or None): The index mode of the database.
        """
        if not hasattr(self._matrix_serializer, "indexer"):
            return None

        return self._matrix_serializer.indexer.index_mode

    def is_deleted(self):
        """
        To check if the database is deleted.

        Returns:
            Bool: True if the database is deleted, False otherwise.
        """
        return self._matrix_serializer.IS_DELETED

    def __len__(self):
        return self.shape[0]

    def __repr__(self):
        return collection_repr(self)

    def __str__(self):
        return self.__repr__()



import json
import shutil

from spinesUtils.asserts import raise_if


class DatabaseManager:
    def __init__(self, root_path):
        self.root_path = root_path
        if not self.root_path.exists():
            self.root_path.mkdir(parents=True, exist_ok=True)

    def register(self, db_name: str):
        raise_if(TypeError, not isinstance(db_name, str), 'db_name must be a string')

        if not (self.root_path / 'databases.json').exists():
            with open(self.root_path / 'databases.json', 'w') as f:
                json.dump([db_name], f)
        else:
            with open(self.root_path / 'databases.json', 'r') as f:
                databases = json.load(f)
                if db_name in databases:
                    return
                databases.append(db_name)
            with open(self.root_path / 'databases.json', 'w') as f:
                json.dump(databases, f)

    def delete(self, db_name: str):
        raise_if(TypeError, not isinstance(db_name, str), 'db_name must be a string')
        if not (self.root_path / 'databases.json').exists():
            return

        databases = self.update_database()

        if db_name in databases:
            databases.remove(db_name)
            shutil.rmtree(self.root_path / db_name)

        with open(self.root_path / 'databases.json', 'w') as f:
            json.dump(databases, f)

    def list_database(self):
        return self.update_database()

    def update_database(self):
        if not (self.root_path / 'databases.json').exists():
            return []

        self.root_path.mkdir(parents=True, exist_ok=True)

        # if a folder does not contain fingerprint.db, exclude it
        folders = [x.name for x in self.root_path.iterdir()
                   if x.is_dir() and (x / '.fingerprint').exists()]

        with open(self.root_path / 'databases.json', 'r') as f:
            databases = json.load(f)

        for db in folders:
            if db not in databases:
                databases.append(db)

        not_existed_db = []
        for db in databases:
            if db not in folders:
                not_existed_db.append(db)

        for db in not_existed_db:
            databases.remove(db)

        with open(self.root_path / 'databases.json', 'w') as f:
            json.dump(databases, f)

        return databases



from .base import BaseIndex
from ..computational_layer.engines import inner_product as ip, l2sq, cosine as cos


class _IndexFlat(BaseIndex):
    def __init__(self):
        """
        Initialize the flat index.
        """
        super().__init__()

    def encode(self, vectors):
        """
        Encode the input data.

        Parameters:
            vectors (np.ndarray): The input data.

        Returns:
            np.ndarray: The encoded data.
        """
        return vectors

    def fit_transform(self, vectors, ids):
        """
        Fit the model and transform the data.

        Parameters:
            vectors (np.ndarray): The input data.
            ids (np.ndarray): The input ids.

        Returns:
            np.ndarray: The encoded data.
        """
        return self.encode(vectors)

    def save(self, filepath):
        """
        Save the model to a file.

        Parameters:
            filepath (str or Pathlike): The name of the file to save the model to.
        """
        ...

    def load(self, filepath):
        """
        Load the model from a file.

        Parameters:
            filepath (str or Pathlike): The name of the file to load the model from.

        Returns:
            _IndexSQ: The loaded model.
        """
        return self


class IndexFlatL2sq(_IndexFlat):
    name = 'IndexFlatL2'

    def __init__(self):
        super().__init__()
        self._register_distance("L2")

    def search(self, original_vec=None, encoded_vec=None, original_data=None,
               encoded_data=None, top_k=10):
        """
        Search for the nearest neighbors of the input data.

        Parameters:
            original_vec (np.ndarray): The original vector.
            encoded_vec (np.ndarray): The encoded vector.
            original_data (np.ndarray): The original data.
            encoded_data (np.ndarray): The encoded data.
            top_k (int): The number of nearest neighbors to return.

        Returns:
            np.ndarray: The indices of the nearest neighbors.
        """
        encoded_vec, encoded_data = super().check_and_encode(original_vec=original_vec, encoded_vec=encoded_vec,
                                                             original_data=original_data, encoded_data=encoded_data)

        return l2sq(encoded_vec, encoded_data, top_k, use_simd=False)


class IndexFlatIP(_IndexFlat):
    name = 'IndexFlatIP'

    def __init__(self):
        super().__init__()
        self._register_distance("IP")

    def search(self, original_vec=None, encoded_vec=None, original_data=None,
               encoded_data=None, top_k=10):
        """
        Search for the nearest neighbors of the input data.

        Parameters:
            original_vec (np.ndarray): The original vector.
            encoded_vec (np.ndarray): The encoded vector.
            original_data (np.ndarray): The original data.
            encoded_data (np.ndarray): The encoded data.
            top_k (int): The number of nearest neighbors to return.

        Returns:
            np.ndarray: The indices of the nearest neighbors.
        """
        encoded_vec, encoded_data = super().check_and_encode(original_vec=original_vec, encoded_vec=encoded_vec,
                                                             original_data=original_data, encoded_data=encoded_data)

        return ip(encoded_vec, encoded_data, top_k, use_simd=False)


class IndexFlatCos(_IndexFlat):
    name = 'IndexFlatCos'

    def __init__(self):
        super().__init__()
        self._register_distance("Cos")

    def search(self, original_vec=None, encoded_vec=None, original_data=None,
               encoded_data=None, top_k=10):
        """
        Search for the nearest neighbors of the input data.

        Parameters:
            original_vec (np.ndarray): The original vector.
            encoded_vec (np.ndarray): The encoded vector.
            original_data (np.ndarray): The original data.
            encoded_data (np.ndarray): The encoded data.
            top_k (int): The number of nearest neighbors to return.

        Returns:
            np.ndarray: The indices of the nearest neighbors.
        """
        encoded_vec, encoded_data = super().check_and_encode(original_vec=original_vec, encoded_vec=encoded_vec,
                                                             original_data=original_data, encoded_data=encoded_data)

        return cos(encoded_vec, encoded_data, top_k, use_simd=False)
